/*
License text attribution: zlib license at http://www.gzip.org/zlib/zlib_license.html
This license is a modification of the zlib license, changing names and dates.

Copyright (C) 2007 Samson Chee

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software
in a product, an acknowledgment in the product documentation would be
appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

Samson Chee samson@herethere.net
*/

#include "stdafx.h"

#include <math.h>
#include <stdio.h>
#include "bubble_sprite.h"

#define kBSBackgroundColor 0x00
#define kBSAmbient 0.5f
#define kBSLightX -1.0f
#define kBSLightY  1.0f
#define kBSLightZ  1.0f

float* bs_calculate_map(int pX, int pY, int pSupersampling)
{
  float theBSLightL = sqrt((kBSLightX * kBSLightX) + (kBSLightY * kBSLightY) + (kBSLightZ * kBSLightZ));

  // create regular map
  float* theRegularMap = new float[pX * pY];
  float* theSuperMap = new float[pX * pY * pSupersampling * pSupersampling];

  for (int i = 0; i < pX * pSupersampling; i++)
  {
    for (int j = 0; j < pY * pSupersampling; j++)
    {
      int theIndex = (j * pX * pSupersampling) + i;

      float theX = (i - ((pX * pSupersampling) / 2.0f)) / ((pX * pSupersampling) / 2.0f);
      float theY = (j - ((pY * pSupersampling) / 2.0f)) / ((pY * pSupersampling) / 2.0f);
      float theLength = sqrt((theX * theX) + (theY * theY));

      if (theLength > 1.0f)
      {
        // outside the circle
        theSuperMap[theIndex] = kBSBackgroundColor;
      }
      else
      {
        // inside the circle
        // - find the unit vector of the sphere at this point
        // - take the cross product of the sphere and the light source direction to get the angle
        // - the lighting at this point is the cosine of the angle + the ambient light

        float theZ = sqrt(1.0f - (theX * theX) - (theY * theY));

        float theDP = (kBSLightX * theX) + (kBSLightY * theY) + (kBSLightZ * theZ);
        float theAngle = acos(theDP / theBSLightL);

        float theDiffuse = cos(theAngle);
        theDiffuse = (theDiffuse < 0) ? 0 : theDiffuse;
        theDiffuse = theDiffuse * theDiffuse;

        float theTotal = (theDiffuse * (1.0f - kBSAmbient)) + kBSAmbient;
        theTotal = (theTotal > 1) ? 1.0f : theTotal;

        theSuperMap[theIndex] = theTotal;
      }

    }
  }

  // reduce from supersampled map to regular map
  for (int i = 0; i < pX; i++)
  {
    for (int j = 0; j < pY; j++)
    {
      float theAcc = 0;

      for (int x = 0; x < pSupersampling; x++)
      {
        for (int y = 0; y < pSupersampling; y++)
        {
          theAcc += theSuperMap[(((j * pSupersampling) + y) * (pX * pSupersampling)) + (i * pSupersampling) + x];
        }
      }

      theAcc /= (float)(pSupersampling * pSupersampling);
      theRegularMap[(j * pX) + i] = theAcc;
    }
  }

  delete theSuperMap;

  return theRegularMap;
}

void bs_free_map(float* pMap)
{
  if (pMap)
  {
    delete pMap;
  }
}

HBITMAP bs_color_sphere(HDC pHDC, int pX, int pY, float* pMap, COLORREF pColor)
{
  BITMAPINFO theBMInfo;
  unsigned char* theBMBits = 0;

  int theX = ((pX + 3) / 4) * 4;
  int theY = pY;

  theBMInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  theBMInfo.bmiHeader.biWidth = theX;
  theBMInfo.bmiHeader.biHeight = theY;
  theBMInfo.bmiHeader.biPlanes = 1;
  theBMInfo.bmiHeader.biBitCount = 24;
  theBMInfo.bmiHeader.biCompression = BI_RGB;
  theBMInfo.bmiHeader.biSizeImage = 0;
  theBMInfo.bmiHeader.biXPelsPerMeter = 0;
  theBMInfo.bmiHeader.biYPelsPerMeter = 0;
  theBMInfo.bmiHeader.biClrUsed = 0;
  theBMInfo.bmiHeader.biClrImportant = 0;

  HBITMAP theBM = CreateDIBSection(pHDC, &theBMInfo, DIB_RGB_COLORS, (void**)&theBMBits, 0, 0);

  for (int i = 0; i < pX; i++)
  {
    for (int j = 0; j < pY; j++)
    {
      int theMapIndex = (j * pX) + i;
      int theBMIndex = ((j * theX) + i) * 3;

      float theMapValue = pMap[theMapIndex];

      theBMBits[theBMIndex + 0] = (unsigned char)(theMapValue * GetBValue(pColor));
      theBMBits[theBMIndex + 1] = (unsigned char)(theMapValue * GetGValue(pColor));
      theBMBits[theBMIndex + 2] = (unsigned char)(theMapValue * GetRValue(pColor));

      //theBMBits[theBMIndex + 0] = 0xff;
      //theBMBits[theBMIndex + 1] = 0xff;
      //theBMBits[theBMIndex + 2] = 0x00;
    }
  }

  return theBM;
}